#!/usr/local/bin/bash
# spinpid.sh for Supermicro boards with single fan zone
VERSION="2017-03-03"
# Run as superuser. See notes at end.

##############################################
#
#  Settings
#
##############################################

# Creates logfile and sends all stdout and stderr to the log, as well as to the console.
# If you want to append to existing log, add '-a' to the tee command.
LOG=/mnt/Ark/Jim/spinpid.log  # Change to your desired log location/name
exec > >(tee -i $LOG) 2>&1

DUTY_MIN=20  # Fan minimum duty cycle (%) (to avoid stalling)

#################  DRIVE SETTINGS ################

SP=33.57        #  Setpoint mean drive temperature (C)

#  Time interval for checking drives (minutes).  Drives change
#  temperature slowly; 5 minutes is probably frequent enough.
T=6
Kp=4    #  Proportional tunable constant (for drives)
Ki=0    #  Integral tunable constant (for drives)
Kd=40   #  Derivative tunable constant (for drives)

#################  CPU SETTINGS ################

#  Time interval for checking CPU (seconds).  1 to 15
#  may be appropriate.
CPU_T=15

#  Reference temperature (C) for scaling CPU_DUTY (NOT a setpoint).
#  At and below this temperature, CPU will demand minimum
#  fan speed (DUTY_MIN above).
CPU_REF=55  # Integer only!
#  Scalar for scaling CPU_DUTY.
#  CPU will demand this number of percentage points in additional
#  duty cycle for each degree of temperature above CPU_REF.
CPU_SCALE=6  # Integer only!

##############################################
# function get_disk_name
# Get disk name from current LINE of DEVLIST
##############################################
# The awk statement works by taking $LINE as input,
# setting '(' as a _F_ield separator and taking the second field it separates
# (ie after the separator), passing that to another awk that uses
# ',' as a separator, and taking the first field (ie before the separator).
# In other words, everything between '(' and ',' is kept.

# camcontrol output for disks on HBA seems to change every version,
# so need 2 options to get ada/da disk name.
function get_disk_name {
   if [[ $LINE == *",p"* ]] ; then     # for ([a]da#,pass#)
      DEVID=$(echo $LINE | awk -F '(' '{print $2}' | awk -F ',' '{print$1}')
   else                                # for (pass#,[a]da#)
      DEVID=$(echo $LINE | awk -F ',' '{print $2}' | awk -F ')' '{print$1}')
   fi
}

############################################################
# function print_header
# Called when script starts and each quarter day
############################################################
function print_header {
   DATE=$(date +"%A, %b %d")
   let "SPACES = DEVCOUNT * 5 + 71"  # 5 spaces per drive
   printf "\n%-*s %-8s %s \n" $SPACES "$DATE" "Fan %" "Interim CPU"
   echo -n "          "
   while read LINE ; do
      get_disk_name
      printf "%-5s" $DEVID
   done <<< "$DEVLIST"             # while statement works on DEVLIST
   printf "%4s %5s %5s %6s %5s %6s %3s %s %4s %-8s %s %s" "Tmax" "Tmean" "ERRc" "P" "I" "D" "CPU" "Driver" "RPM" "MODE" "Curr/New" "Adjustments"
}

#################################################
# function read_fan_data
#################################################
function read_fan_data {
   # Read duty cycle
   DUTY_CURR=$($IPMITOOL raw 0x30 0x70 0x66 0 0) # in hex with leading space
   DUTY_CURR=$((0x$(echo $DUTY_CURR)))  # echoing trims leading space, then decimalize
   # Read fan mode
   MODE=$($IPMITOOL raw 0x30 0x45 0) # in hex
   MODE=$((0x$(echo $MODE)))  # strip leading space and decimalize
   # Text for mode
   case $MODE in
      0) MODEt="Standard" ;;
      4) MODEt="HeavyIO" ;;
      2) MODEt="Optimal" ;;
      1) MODEt="Full" ;;
   esac
   # Get reported fan speed in RPM.
   # Takes the line with FAN1, then 2nd through the 5th
   # digit if there are that many.
   RPM=$($IPMITOOL sdr | grep "FAN1" | grep -Eo '[0-9]{2,5}')
}

##############################################
# function CPU_check_adjust
# Get CPU temp. Calculate a new DUTY_CPU.
# If it is greater than the duty due to the
# drives, send it to adjust_fans.
##############################################
function CPU_check_adjust {
   CPU_TEMP=$($IPMITOOL sdr | grep "CPU Temp" | grep -Eo '[0-9]{2,5}')
   # This will break if settings have non-integers
   let DUTY_CPU=(CPU_TEMP-CPU_REF)*CPU_SCALE+DUTY_MIN
   
   if [ $FIRST_TIME -eq 1 ]; then return; fi
   
   MAX=$(( DUTY_DRIVE > DUTY_CPU ? DUTY_DRIVE : DUTY_CPU ))

   # DIAGNOSTIC variables - uncomment for troubleshooting:
   # printf "\nDUTY_DRIVE=%s, CPU_TEMP=%s, DUTY_CPU=%s, MAX=%s, DUTY_CURR=%s  " "${DUTY_DRIVE:---}" "${CPU_TEMP:---}" "${DUTY_CPU:---}" "${MAX:---}" "${DUTY_CURR:---}"

   adjust_fans $MAX  # passing higher duty to the function adjust_fans

   # DIAGNOSTIC variables - uncomment for troubleshooting:
   # printf ", DUTY_NEW=%s" "${DUTY_NEW:---}"

   if [[ DUTY_NEW -ne DUTY_CURR ]]; then
      DUTY_CURR=$DUTY_NEW
      printf "%d " $DUTY_NEW
   fi
   if [[ $DUTY_CPU -gt $DUTY_DRIVE ]]; then
      # MAX may be modified by adjust_fans as DUTY_NEW
      DUTY_CPU=$DUTY_NEW
   fi

   # DIAGNOSTIC variables - uncomment for troubleshooting:
   # printf " UPDATED DUTY_CPU=%s, DUTY_CURR=%s" $DUTY_CPU $DUTY_CURR
}

##############################################
# function DRIVES_check_adjust
# Print time on new log line.
# Go through each drive, getting and printing
# status and temp.  Calculate sum and max
# temp, then call function drive_data.
# Apply max of $PID and CPU_CORR to the fans.
##############################################
function DRIVES_check_adjust {
   echo  # start new line
   # print time on each line
   TIME=$(date "+%H:%M:%S"); echo -n "$TIME  "
   Tmax=0; Tsum=0  # initialize drive temps for new loop through drives
   i=0  # count number of spinning drives
   while read LINE ; do
      get_disk_name
      /usr/local/sbin/smartctl -a -n standby "/dev/$DEVID" > /var/tempfile
      RETURN=$?  # have to preserve return value or it changes
      BIT0=$(($RETURN & 1))
      BIT1=$(($RETURN & 2))
      if [ $BIT0 -eq 0 ]; then
         if [ $BIT1 -eq 0 ]; then  # spinning
            STATUS="*"
         else  # drive found but no response, probably standby
            STATUS="_"
         fi
      else   # smartctl returns 1 (00000001) for missing drive
         STATUS="?"
      fi

      TEMP=""
      # Update temperatures each drive; spinners only
      if [ "$STATUS" == "*" ] ; then
         # Taking 10th space-delimited field for WD, Seagate, Toshiba, Hitachi
         TEMP=$( cat /var/tempfile | grep "Temperature_Celsius" | awk '{print $10}')
         let "Tsum += $TEMP"
         if [[ $TEMP > $Tmax ]]; then Tmax=$TEMP; fi;
         let "i += 1"
      fi
      printf "%s%-2d  " "$STATUS" $TEMP
   done <<< "$DEVLIST"

   # if no fans are spinning
   if [ $i -eq 0 ]; then
      Tmean=""; Tmax=""; P=""; D=""
      DUTY_DRIVE=$DUTY_MIN
   else
      # summarize, calculate PID and print Tmax and Tmean
      Tmean=$(echo "scale=3; $Tsum / $i" | bc)
      ERRp=$ERRc
      ERRc=$(echo "scale=3; ($Tmean - $SP) / 1" | bc)
      ERR=$(echo "$ERRc * $T + $I" | bc)
      P=$(echo "scale=3; ($Kp * $ERRc) / 1" | bc)
      P=$(printf %0.2f $P) # add leading 0 if needed, 2 dec. places
      I=$(echo "scale=2; ($Ki * $ERR) / 1" | bc)
      D=$(echo "scale=3; $Kd * ($ERRc - $ERRp) / $T" | bc)
      D=$(printf %0.2f $D) # add leading 0 if needed, 2 dec. places
      PID=$(echo "$P + $I + $D" | bc)  # add 3 corrections
      PID=$(printf %0.f $PID)  # round

      # add leading 0 if needed, round for printing
      Tmean=$(printf %0.2f $Tmean)
      ERRc=$(printf %0.2f $ERRc)
      P=$(printf %0.2f $P)
      D=$(printf %0.2f $D)
      PID=$(printf %0.f $PID)

      let "DUTY_DRIVE = $DUTY_DRIVE + $PID"
   fi

   # print current Tmax, Tmean
   printf "^%-3s %5s" "${Tmax:---}" "${Tmean:----}"

   if [[ $DUTY_DRIVE -ge $DUTY_CPU ]]; then
      DRIVER="Drives"
      adjust_fans $DUTY_DRIVE
      DUTY_DRIVE=$DUTY_NEW
   else
      DRIVER="CPU"
      adjust_fans $DUTY_CPU
   fi
}

##############################################
# function adjust_fans
# Constrain passed duty, then set if different
##############################################
function adjust_fans {
   DUTY_NEW=$1
   
# Don't allow duty cycle beyond $DUTY_MIN/95%
   if [[ $DUTY_NEW -gt 95 ]]; then DUTY_NEW=95; fi
   if [[ $DUTY_NEW -lt $DUTY_MIN ]]; then DUTY_NEW=$DUTY_MIN; fi
   # Change if different from current duty
   if [[ $DUTY_NEW -ne $DUTY_CURR ]]; then
      # Set new duty cycle. "echo -n ``" prevents newline generated in log
      echo -n `$IPMITOOL raw 0x30 0x70 0x66 1 0 $DUTY_NEW`
   fi
}

#####################################################
# All this happens only at the beginning
# Initializing values, list of drives, print header
#####################################################
CPU_LOOPS=$( echo "$T * 60 / $CPU_T" | bc )  # Number of whole CPU loops per drive loop
IPMITOOL=/usr/local/bin/ipmitool
I=0; ERRc=0  # Initialize errors to 0
FIRST_TIME=1

# Get list of drives
DEVLIST1=$(/sbin/camcontrol devlist)
# Remove lines with flash drives or SSD; edit as needed
# You could use another strategy, e.g., find something in the camcontrol devlist
# output that is unique to the drives you want, for instance only WDC drives:
# if [[ $LINE != *"WDC"* ]] . . .
DEVLIST="$(echo "$DEVLIST1"|sed '/KINGSTON/d;/ADATA/d;/SanDisk/d;/OCZ/d;/LSI/d;/INTEL/d;/TDKMedia/d')"
DEVCOUNT=$(echo "$DEVLIST" | wc -l)

read_fan_data   # get fan status before making any adjustments

# If mode not full, set it to avoid BMC changing duty cycle
# Need to wait a tick or it may not get next command
# "echo -n" to avoid annoying newline generated in log
if [[ MODE -ne 1 ]]; then
   echo -n `$IPMITOOL raw 0x30 0x45 1 1`; sleep 1
   echo -n `$IPMITOOL raw 0x30 0x70 0x66 1 0 50`; sleep 1
fi

# DUTY_DRIVE NEEDS initial value.  Use DUTY_CURR unless it is
# very high and would take a long time to equilibrate.
# (or doesn't exist; second test true if it exists)
if [[ $DUTY_CURR -lt 50 && -n ${DUTY_CURR+x} ]]; then
   DUTY_DRIVE=$DUTY_CURR
else
   DUTY_DRIVE=50
fi

# Before starting, go through the drives to report if
# smartctl return value indicates a problem (>2).
# Use -a so that all return values are available.
while read LINE ; do
   get_disk_name
   /usr/local/sbin/smartctl -a -n standby "/dev/$DEVID" > /var/tempfile
   if [ $? -gt 2 ]; then
      printf "\n"
      printf "*******************************************************\n"
      printf "* WARNING - Drive %-4s has a record of past errors,   *\n" $DEVID
      printf "* is currently failing, or is not communicating well. *\n"
      printf "* Use smartctl to examine the condition of this drive *\n"
      printf "* and conduct tests. Status symbol for the drive may  *\n"
      printf "* be incorrect (but probably not).                    *\n"
      printf "*******************************************************\n"
   fi
done <<< "$DEVLIST"

printf "\n%s %36s %s \n" "Key to drive status symbols:  * spinning;  _ standby;  ? unknown" "Version" $VERSION
print_header
CPU_check_adjust

###########################################
# Main loop through drives every T minutes
# and CPU every CPU_T seconds
###########################################
while [ 1 ] ; do
   # Print header every quarter day.  Expression removes any
   # leading 0 so it is not seen as octal
   HM=$(date +%k%M); HM=`expr $HM + 0`
   R=$(( HM % 600 ))  # remainder after dividing by 6 hours
   if (( $R < $T )); then
      print_header;
   fi

   if [[ $FIRST_TIME -eq 0 ]]; then read_fan_data; fi
   FIRST_TIME=0

   DRIVES_check_adjust

   printf "%6s %6s %5s %6.6s %3d %-6s %4d %-8s %2d/%-6d" "${ERRc:----}" "${P:----}" $I "${D:----}" $CPU_TEMP $DRIVER $RPM $MODEt $DUTY_CURR $DUTY_NEW
   DUTY_CURR=$DUTY_NEW

   i=0
   while [ $i -lt $CPU_LOOPS ]; do
      CPU_check_adjust
      sleep $CPU_T
      let i=i+1
   done
done
